"""
Export TextGrid files to other formats
"""
import csv
import json
from decimal import Decimal
from jinja2 import Environment, PackageLoader, select_autoescape
from pathlib import Path

def textgrid_to_text_file(textgrid_obj, dst_path, short_format = False, encoding = 'utf-8'):
    """
    Write a TextGrid object to a text file.

    Parameters
    ----------
    textgrid_obj : 
        A TextGrid object.
    dst_path : str or :class:`pathlib.Path`
        The path where the text file will be stored.
    short_format : boolean, default `False`
        True for short format of the text file. Otherwirse, long format is applied.
    encoding: str, default 'utf-8'
        The encoding of the text file.
    """
    env = Environment(
        loader=PackageLoader('mytextgrid.io'),
        autoescape=select_autoescape()
    )
    env.trim_blocks = True
    env.lstrip_blocks  = True

    # Variales
    dict_ = textgrid_obj.to_dict()
    template_name = 'short_format.TextGrid.jinja'if short_format else 'long_format.TextGrid.jinja'
    template = env.get_template(template_name)
    textgrid_str = template.render(textgrid = dict_)

    with open(dst_path, 'w', encoding = encoding) as textfile:
        textfile.write(textgrid_str)

def to_csv(textgrid_obj, path, encoding = 'utf-8'):
    """
    Write TextGrid to a csv file.

    Parameters
    ----------
    textgrid_obj : TextGrid
        Data stored in a TextGrid.
    path : str
        The path where the delimited text file will be created.
    encoding : str, default utf-8
        The encoding of the resulting file.

    Notes
    -----
    By using this function, `TextGrid` is converted to a table where non-empty items (`Interval`
    and `Point`) become rows. Each item (row) is stored along with the tier it belongs to and
    its timestamps. When the table is done, the items (rows) are sorted by time in ascending
    order. Finally, the table is exported as a delimited text file.
    """
    table = []
    for tier in textgrid_obj:
        for item in tier:
            if item.text == '':
                continue
            if tier.is_interval():
                # IntervalTier class
                table.append([item.xmin, tier.name, item.text, item.xmax])
            else:
                # PointTier class
                table.append([item.time, tier.name, item.text, item.time])
    table.sort(key=lambda x:x[0])

    with open(path, 'w', encoding = encoding, newline = '') as file_object:
        spamwriter = csv.writer(file_object)
        spamwriter.writerow(['tmin', 'tier_name', 'text', 'tmax'])
        for row in table:
            spamwriter.writerow([str(item) for item in row])

def textgrid_to_json(textgrid_obj, path, encoding = 'utf-8'):
    """
    Write TextGrid to a json file.

    Parameters
    ----------
    textgrid_obj : TextGrid
        Data stored in a TextGrid.
    path : str
        The path where the delimited text file will be created.
    encoding : str, default utf-8
        The encoding of the resulting file.
    """
    dict_ = textgrid_obj.to_dict()
    with open(path, 'w', encoding = encoding) as file_object:
        json.dump(dict_, file_object, cls = _DecimalEncoder, ensure_ascii = False, indent = 4)

class _DecimalEncoder(json.JSONEncoder):
    """
    JSON encoder for TextGrid.
    """
    def default(self, decimal_number):
        if isinstance(decimal_number, Decimal):
            return str(decimal_number)
        else:
            return super().default(decimal_number)



"""Create and manipulate `Interval` and `IntervalTier` objects."""
import decimal
from mytextgrid.core.tier import Tier
from mytextgrid.eval import obj_to_decimal
decimal.getcontext().prec = 16

class IntervalTier(Tier):
    """
    A class representation for an interval tier.
    """
    def __init__(self, name = '', xmin = 0, xmax = 1):
        """
        Initialize an instance of :class:`~mytextgrid.core.interval_tier.IntervalTier`.

        Parameters
        ----------
        name : str, default ''
            The name of the tier.
        xmin : int, float str or :class:`decimal.Decimal`
            The starting time (in seconds) of the tier.
        xmin : int, float str or :class:`decimal.Decimal`
            The ending time (in seconds) of the tier.
        """
        super().__init__(name, xmin, xmax, is_interval = True)
        self._items = [Interval(self._xmin, self._xmax)]

    def insert_boundaries(self, *times):
        """
        Insert one or more time boundaries into a
        :class:`~mytextgrid.core.interval_tier.IntervalTier`.

        Parameters
        ----------
        *time : float or decimal.Decimal
            The times (in seconds) of the new boundaries.
        """
        for time in times:
            self.insert_boundary(time)

    def insert_boundary(self, time):
        """
        Insert a time boundary into a :class:`~mytextgrid.core.interval_tier.IntervalTier`.

        Parameters
        ----------
        time : int, float, str or :clas::`decimal.Decimal`
            The time (in seconds) of the new boundary.

        Returns
        -------
        tuple of (int, int)
            The left and right interval position at the inserted boundary.

        Raises
        ------
        ValueError
            If the specified time already exists.
        """
        time_ = obj_to_decimal(time)
        self.eval_time_range(time_) # Raise exceptions if out of the tier range

        index = self.get_index_at_time(time_)
        old_interval = self._items[index]

        if old_interval.xmin == time_:
            raise ValueError(f'There is already a boundary at {time_}')

        new_left_interval = Interval(
            xmin = old_interval.xmin,
            xmax = time_,
            text = old_interval.text
        )

        new_right_interval = Interval(
            xmin = time_,
            xmax = old_interval.xmax
        )

        # Update items
        self._items[index] = new_left_interval
        self._items.insert(index+1, new_right_interval)

        # Returns
        return (index, index+1)

    def remove_boundary(self, time):
        """
        Remove a time boundary from a :class:`~mytextgrid.core.interval_tier.IntervalTier`.

        Parameters
        ----------
        time : int, float, str or :clas::`decimal.Decimal`
            The time (in seconds) of the new boundary.

        Raises
        ------
        ValueError
            If there is there is not a boundary at the specified time.
        """
        time_ = obj_to_decimal(time)
        self.eval_time_range(time_) # Raise exceptions if out of range

        index = self.get_index_at_time_boundary(time_)
        if index is None:
            raise ValueError(f'No boundary to remove at {time}.')

        interval_left = self._items[index-1]
        interval_right = self._items[index]

        # Create and insert new interval
        new_interval  = Interval(
            xmin = interval_left.xmin,
            xmax = interval_right.xmax,
            text = interval_left.text + interval_right.text
        )

        # Update items
        self._items[index] = new_interval
        self._items.remove(interval_left)

    def move_boundary(self, src_time, dst_time):
        """
        Move a boundary to another location between the time range of the left and right intervals
        of the source location.

        Parameters
        ----------
        src_time : float or :class:`decimal.Decimal`
            The time (in seconds) of an existing boundary.
        dst_time : float or :class:`decimal.Decimal`
            The time (in seconds) where the boundary will be moved to.

        Raises
        ------
        ValueError
            If the new time location is outside of the time range of the left and right intervals.
        """
        # Normalize numbers
        src_time_ = obj_to_decimal(src_time)
        dst_time_ = obj_to_decimal(dst_time)

        # Raise exceptions.
        if src_time_ == dst_time_:
            raise ValueError('src_time and dst_time MUST NOT BE equal')
        self.eval_time_range(src_time_)
        self.eval_time_range(dst_time_)

        ## Get/Check if src boundary time exists.
        index = self.get_index_at_time_boundary(src_time_)
        if index is None:
            raise ValueError(f'No boundary found at src_time ({src_time_})')

        left_interval = self._items[index-1]
        right_interval = self._items[index]

        ## Check if dst_time is outside its neighbors boundaries.
        if left_interval.xmin >= dst_time_ >= right_interval.xmax:
            raise ValueError('Cannot move the source boundary outside its neighbors boundaries.')

        # Create intervals objects and replace them.
        new_left_interval = Interval(
            xmin = left_interval.xmin,
            xmax = dst_time_,
            text = left_interval.text
        )

        new_right_interval = Interval(
            xmin = dst_time_,
            xmax = right_interval.xmax,
            text = right_interval.text
        )

        # Update items
        self._items[index-1] = new_left_interval
        self._items[index] = new_right_interval

    def set_text_at_index(self, index, *text_items):
        """
        Set the text of one or more of intervals.

        If more than one text item is provided, intervals will be set from left to right
        counting from the starting `index`.

        Parameters
        ----------
        index : int
            The index at which text items will be inserted.
        text_items : str or iterable of str
            The text items that will be inserted.
        """
        # Raise Exception if more text items than intervals.
        if (index + len(text_items) - 1) > len(self._items):
            raise IndexError('more text items than intervals.')

        for index_, text in enumerate(text_items, start = index):
            self._items[index_].text = text

    def get_index_at_time(self, time):
        """
        Return the index of an interval at the given time.

        If time relies on the boundary between two intervals, it returns
        the index of the right interval.

        Parameters
        ----------
        time : int, float, str or :class:`decimal.Decimal`
            The time in seconds to be evaluated.

        Returns
        -------
        int or None
            Return the interval index at the specified time.
        """
        time_ = obj_to_decimal(time)

        if time_ == self._xmax:
            return len(self._items) - 1 # Returns the last index

        for index, interval in enumerate(self._items):
            if interval.xmin <= time_ < interval.xmax:
                return index
        return None

    def get_interval_at_time(self, time):
        """
        Return the interval at the specified time.

        If `time` relies on the boundary between two intervals, it returns
        the the right interval.

        Parameters
        ----------
        time : int, float, str or :class:`decimal.Decimal`
            The time in seconds to be evaluated.

        Returns
        -------
        :class:`mytextgrid.core.interval_tier.Interval` or None
            Return the interval index at the specified time.
        """
        index = self.get_index_at_time(time)
        if index is None:
            return None
        return self._items[index]

    def get_index_at_time_boundary(self, time):
        """
        Get the index of an interval at a given boundary.

        Only counts the inserted boundaries.

        Parameters
        ----------
        time : int, float, str or :class:`decimal.Decimal`
            The time to match with any boundary.

        Returns
        -------
        int or None
            Return an integer if the
        """
        time_ = obj_to_decimal(time)

        for index, interval in enumerate(self._items):
            if index == 0:
                continue
            if interval.xmin == time_:
                return index
        return None

class Interval:
    """
    A class representation for an interval.
    """
    def __init__(self, xmin, xmax, text = ''):
        """
        Init an object representing a Praat interval.

        Parameters
        ----------
        xmin : int, float str or decimal.Decimal
            The starting time (in seconds) of the interval.
        xmin : int, float str or decimal.Decimal
            The ending time (in seconds) of the interval.
        text : str, default ''
            The text content of the interval.
        """
        # Check input type
        if not isinstance(text, str):
            raise TypeError('text MUST BE a str.')
        if not isinstance(xmin, (int, float, str, decimal.Decimal)):
            raise TypeError('xmin MUST BE an int, float, str or decimal.Decimal.')
        if not isinstance(xmax, (int, float, str, decimal.Decimal)):
            raise TypeError('xmax MUST BE an int, float, str or decimal.Decimal.')

        xmin_ = obj_to_decimal(xmin)
        xmax_ = obj_to_decimal(xmax)

        assert xmax_ > xmin_, 'xmax MUST BE greater than xmin'

        # Assign attributes
        self._xmin = xmin_
        self._xmax = xmax_
        self._text = text

    @property
    def xmin(self):
        """
        Return _xmin attribute.
        """
        return self._xmin

    @property
    def xmax(self):
        """
        Return _xmax attribute.
        """
        return self._xmax

    @property
    def text(self):
        """
        Return _text attribute.
        """
        return self._text

    @text.setter
    def text(self, value):
        """
        Modify _text attribute.
        """
        if not isinstance(value, str):
            raise TypeError('text MUST BE a str')
        self._text = value

    def get_duration(self):
        """
        Return the duration of the interval.

        Return
        ------
        :class:`decimal.Decimal`
            The duration of the interval.
        """
        return self._xmax - self._xmin



"""Parse TextGrid files in full text format into TextGrid objects"""
import decimal
import re
from io import StringIO
from pathlib import Path
import chardet
from mytextgrid.core.textgrid import TextGrid
decimal.getcontext().prec = 16

def parse_textgrid_file(path, encoding = None):
    """
    Parse a full text TextGrid file into a dict.

    Parameters
    ----------
    path : str or :class:`pathlib.Path`
        The path of the TextGrid file.
    encoding : str or None, default None
        The encoding of the TextGrid file. If None, the function will figure out
        the encoding.
    """
    if isinstance(path, str):
        path = Path(path)

    # Open file
    if encoding is None:
        encoding = _detect_encoding(path)

    with open(path, 'r', encoding = encoding) as file_object:
        dict_ = parse(file_object)
        dict_['basename'] = path.stem
        dict_['path'] = path
    return dict_

def parse(stream, name = None, path = None):
    """
    Parse a full text TextGrid file into a dict.

    {
        'basename': str,
        'path': :class:`pathlib.Path`,
        'xmin': str,
        'xmax': str,
        'tiers': [
            {
                class: str,
                tier_name: str,
                items: [
                    {
                        'xmin': str
                        'xmax': str
                        '{"text", "mark"}': str
                    },
                    .
                    .
                    .
                ]
            },
            .
            .
            .
        ]
    }

    Parameters
    ----------
    stream : str or :class:`io.StringIO`
        The content of a full-formatted TextGrid.
    name : str
        The name of the TextGrid
    path : str or :class:`pathlib.Path`
        The path of the TextGrid

    Returns
    -------
        dict
            A `dict` representation of the TextGrid file.
    """
    if isinstance(stream, str):
        stream = StringIO(stream)
    if isinstance(path, str):
        path = Path(path)

    textgrid = {
        'basename': name,
        'path': path,
        'xmin': None,
        'xmax': None,
        'tiers':[]
    }

    line = stream.readline()
    while line:
        key, match = _parse_line(line)
        # Check header
        if key == 'file_type':
            file_type = match.group(key)
            if not file_type == 'ooTextFile':
                raise OSError('The stream is not a Praat object.')

        if key == 'object_class':
            object_class = match.group(key)
            if not object_class == 'TextGrid':
                raise OSError('The stream is not a TextGrid.')

        # TextGrid info
        if key == 'tg_xmin':
            textgrid['xmin'] = match.group(key)

        if key == 'tg_xmax':
            textgrid['xmax'] = match.group(key)

        # Tier info
        if key == 'tier_class':
            textgrid['tiers'].append(
                {
                'class': match.group(key),
                'tier_name':None,
                'items': []
                }
            )

        if key == 'tier_name':
            textgrid['tiers'][-1]['tier_name'] = match.group(key)

        # Item content
        if key == 'interval_xmin':
            textgrid['tiers'][-1]['items'].append(
                {
                'xmin': match.group(key),
                'xmax': None,
                'text': None
                }
            )

        if key == 'interval_xmax':
            textgrid['tiers'][-1]['items'][-1]['xmax'] = match.group(key)

        if key == 'point_number':
            textgrid['tiers'][-1]['items'].append(
                {'number':match.group(key),
                'mark':None
                }
            )

        if key in ('interval_text', 'point_mark'):
            type_label = 'text' if key == 'interval_text' else 'mark'
            text = match.group(key).replace('""', '"')
            textgrid['tiers'][-1]['items'][-1][type_label] = text

        # If the text or mark field has multiple lines, read the those
        # lines until the last.
        if key in ('interval_text2', 'point_mark2'):
            text = match.group(key) + '\n'
            enditem_pattern = re.compile('(.*)" $')
            type_label = 'text' if key == 'interval_text2' else 'mark'
            item = textgrid['tiers'][-1]['items'][-1]

            while line:
                line = stream.readline()
                match = enditem_pattern.match(line)
                if match:
                    # Check the last character not to be a `"" \n`
                    if not line.endswith('"" \n'):
                        text = text + match.group(1)
                        item[type_label] = text.replace('""', '"')
                        break
                text = text + line
        line = stream.readline()
    return textgrid

def dict_to_textgrid(textgrid):
    """
    Build and return a :clas:`mytextgrid.TextGrid` from a dict formatted TextGrid.

    Parameters
    ----------
    textgrid : dict of a TextGrid
        A dict formatted created by :meth:`FullTextParser.full_textgrid_to_dict`.

    Returns
    -------
        :class:`mytextgrid.TextGrid`
            A TextGrid object.

    See also
    --------
    mytextgrid.Parser.full_textgrid_to_dict : Returns a dict formatted str created from a
    TextGrid file.
    """
    # Init TextGrid object
    textgrid_obj = TextGrid(
        xmin = decimal.Decimal(textgrid['xmin']),
        xmax = decimal.Decimal(textgrid['xmax'])
        )

    for tier in textgrid['tiers']:
        tier_class = tier['class']
        tier_name = tier['tier_name']

        if tier_class == 'IntervalTier':
            # Insert interval tier
            tier_obj = textgrid_obj.insert_tier(tier_name)

            # Insert interval
            for loc, item in enumerate(tier['items']):
                # Insert boundary.
                if not loc == 0:
                    tier_obj.insert_boundaries(
                        decimal.Decimal(item['xmin'])
                    )
                # Insert text
                tier_obj.set_text_at_index(loc, item['text'])

        if tier_class == 'TextTier':
            # Insert tier
            tier_obj = textgrid_obj.insert_tier(tier_name, False)
            for item in tier['items']:
                # Insert Point
                tier_obj.insert_point(
                    decimal.Decimal(item['number']),
                    item['mark']
                )
    return textgrid_obj

def _parse_line(line):
    """
    Test if a list of patterns stored in dictonary match a text line.

    Parameters
    ----------
    line : str
        Any line in full textgrid format that comes from a TextGrid file.

    Returns
    -------
        (key, match), tuple of (str, re.Match)
            Return the key of the pattern that matches the line and the `re.Match`.
            If no key matches the line, return (None, None)
    """
    _patterns_dict = {
        'file_type': re.compile('File type = "(?P<file_type>.+)"'),
        'object_class': re.compile('Object class = "(?P<object_class>.+)"'),
        'tg_xmin': re.compile('xmin = (?P<tg_xmin>.+) '),
        'tg_xmax': re.compile('xmax = (?P<tg_xmax>.+) '),
        'tier_exists': re.compile(r'tiers? <(?P<tier_exists>.+)> '),
        'tiers': re.compile(r'size = (?P<tiers>\d+) '),
        'tier_loc': re.compile(r' {4}item [(?P<tier_loc>\d+)]'),
        'tier_class': re.compile(' {8}class = "(?P<tier_class>.+)" '),
        'tier_name': re.compile(' {8}name = "(?P<tier_name>.*)" '),
        'tier_xmin': re.compile(' {8}xmin = (?P<tier_xmin>.+) '),
        'tier_xmax': re.compile(' {8}xmax = (?P<tier_xmax>.+) '),
        'intervals': re.compile(r' {8}intervals: size = (?P<intervals>\d+) '),
        'interval_loc': re.compile(r' {8}intervals [(?P<interval_loc>\d+)]:'),
        'interval_xmin': re.compile(' {12}xmin = (?P<interval_xmin>.+) '),
        'interval_xmax': re.compile(' {12}xmax = (?P<interval_xmax>.+) '),
        'interval_text': re.compile(' {12}text = "(?P<interval_text>.*)" '),
        'interval_text2': re.compile(' {12}text = "(?P<interval_text2>.*)'),
        'points': re.compile(r' {8}points: size = (?P<points> \d+) '),
        'point_loc': re.compile(r' {8}points [(?P<point_loc>\d+)]:'),
        'point_number': re.compile(' {12}number = (?P<point_number>.+) '),
        'point_mark': re.compile(' {12}mark = "(?P<point_mark>.*)" '),
        'point_mark2': re.compile(' {12}mark = "(?P<point_mark2>.*)')
    }
    for key, pattern in _patterns_dict.items():
        match = pattern.match(line)
        if match:
            return key, match
    return None, None

def _detect_encoding(path):
    """
    Detect and return a file encoding.

    Parameters
    ----------
    path : str
        The path of a text file.

    Returns
    -------
        str
            The detected encoding name of the specified file.
    """
    with open(path, 'rb') as file_object:
        rawdata = file_object.read()
        result = chardet.detect(rawdata)
        return result.get('encoding')



from mytextgrid.io.parser import fulltext_format

def read_from_file(path, encoding = None):
    """
    Read a TextGrid file with full text format and return a TextGrid object.

    Parameters
    ----------
    path : str
        The path of the TextGrid file.
    encoding : str, default None, detect automatically the encoding.
        The name of the encoding used to decode the file. See the codecs module for the list
        supported encodings.

    Returns
    -------
        :class:`mytextgrid.TextGrid`
            A TextGrid instance.
    """
    textgrid_dict = fulltext_format.parse_textgrid_file(path, encoding)
    textgrid_obj = fulltext_format.dict_to_textgrid(textgrid_dict)
    return textgrid_obj

def read_from_stream(*args, **kwds):
    """
    Read a stream into a TextGrid object.

    Parameters
    ----------
    stream : str or :class:`io.StringIO`
        The content of a full-formatted TextGrid.
    name : str
        The name of the TextGrid.
    path : str or :class:`pathlib.Path`
        The path of the TextGrid.

    Returns
    -------
        :class:`mytextgrid.TextGrid`
            A TextGrid instance.
            A TextGrid instance.
    """
    textgrid_dict = fulltext_format.parse(*args, **kwds)
    textgrid_obj = fulltext_format.dict_to_textgrid(textgrid_dict)
    return textgrid_obj



"""
Create and manipulate TextGrid objects.
"""
import decimal
from mytextgrid.io import textgrid_out
from mytextgrid.core.interval_tier import IntervalTier
from mytextgrid.core.point_tier import PointTier
from mytextgrid.eval import obj_to_decimal
decimal.getcontext().prec = 16

def create_textgrid(xmin = 0, xmax = 1):
    """
    Create and return an empty TextGrid.

    An empty TextGrid does not contain any tier.

    Parameters
    ----------
    xmin : int, float, str or :class:`decimal.Decimal`, default 0
        The starting time of the TextGrid.
    xmax : int, float, str or :class:`decimal.Decimal`, default 1
        The ending time of the TextGrid.

    Returns
    -------
        :class:`mytextgrid.TextGrid`
            A TextGrid instance.
    """
    return TextGrid(xmin, xmax)

class TextGrid:
    """
    A class representation for a TextGrid.
    """
    def __init__(self, xmin = 0, xmax = 1):
        if not isinstance(xmin, (int, float, str, decimal.Decimal)):
            raise TypeError('xmin MUST BE an int, float, str or decimal.Decimal.')
        if not isinstance(xmax, (int, float, str, decimal.Decimal)):
            raise TypeError('xmax MUST BE an int, float, str or decimal.Decimal.')

        xmin_ = obj_to_decimal(xmin)
        xmax_ = obj_to_decimal(xmax)
        assert xmax_ > xmin_, 'xmax MUST BE greater than xmin'

        # Attributes
        self._xmin = xmin_
        self._xmax = xmax_
        self._tiers = []

    @property
    def xmin(self):
        """
        Return the `_xmin` attribute.
        """
        return self._xmin

    @property
    def xmax(self):
        """
        Return the `_xmax` attribute.
        """
        return self._xmax

    @property
    def tiers(self):
        """
        Return the `_tiers` attribute.
        """
        return self._tiers

    def __len__(self):
        return len(self._tiers)

    def __iter__(self):
        return iter(self._tiers)

    def __getitem__(self, key):
        return self._tiers[key]

    def describe(self):
        """
        Show the attributes and structure of a TextGrid.

        Show tier information: index, type, name and size.

        Examples
        --------
        >>> tg = mytextgrid.create_textgrid(0, 1)
        >>> tone_tier = tg.insert_tier("tone", False)
        >>> tone_tier.insert_tier('tone', 0.66, "H")
        >>> tone_tier.insert_tier('tone', 0.9, "L")
        >>> segment_tier = tg.insert_tier("segment")
        >>> segment_tier.insert_boundaries('segment', 0.23, 0.30, 0.42, 0.62, 0.70, 0.82, 0.98)
        >>> segment_tier.set_interval_text('segment', 1, 'e', 'l', 'p', 'e', 'rr', 'o')

        Use :meth:`~mytextgrid.describe` to describe a TextGrid.

        >>> tg.describe()
            TextGrid:
                Name:                  perro
                Startig time (sec):    0
                Ending time (sec):     1
                Number of tiers:       2
            Tiers Summary:
            0	TextTier	tone	(size = 2)
            1	IntervalTier	segment	(size = 8)
        """
        size = len(self._tiers)

        summary = []
        for index, tier in enumerate(self._tiers):
            tier_class = 'IntervalTier' if tier.is_interval() else 'PointTier'
            summary.append(
                f'    {index}\t'
                f'{tier_class}\t'
                f'{tier.name}\t'
                f'(size = {len(tier)})'
                )

        summary_str = ''
        if size > 0:
            content = '\n'.join(summary)
            head = "Tiers summary:"
            summary_str = f'\n{head}\n{content}'

        message = (
            'TextGrid:\n'
            f'    Startig time (sec):    {self.xmin}\n'
            f'    Ending time (sec):     {self.xmax}\n'
            f'    Number of tiers:       {size}'
            f'{summary_str}'
            )
        print(message)

    def get_duration(self):
        """
        Return the duration of the TextGrid in seconds.

        Returns
        -------
        :class:`decimal.Decimal`
            Time duration in seconds.

        Examples
        --------
        >>> tg = mytextgrid.create_textgrid('banana', 0, 1.2)
        >>> tg.get_duration()
            1.2
        """
        return self._xmax - self._xmin

    def insert_tier(self, name, interval_tier = True, index = None):
        """
        Insert a tier in the TextGrid.

        Parameters
        ----------
        name : str
            The name of the tier.
        interval_tier : bool
            If True, insert an :class:`IntervalTier`. Otherwise, insert a :class:`PointTier`.
        index : int, default None, meaning the last index.
            The index of the tier.

        Examples
        --------
        >>> tg = mytextgrid.create_textgrid('banana', 0, 1.2)

        Insert an interval tier.

        >>> tg.insert_tier('word')

        To create a point tier, set ``interval_tier`` as ``False``.

        >>> tg.insert_tier('tone', interval_tier = False)

        With ``index``, you can insert a tier at a specific position.

        >>> tg.insert_tier('phrase', index = 0)

        Returns
        -------
        :class:`IntervalTier` or :class:`PointTier`
            An empty tier.
        """
        if index is None:
            index = len(self)

        if not isinstance(index, int):
            raise ValueError('index MUST BE a int value.')

        if interval_tier:
            tier = IntervalTier(name, self._xmin, self._xmax)
        else:
            tier = PointTier(name, self._xmin, self._xmax)
        self.tiers.insert(index, tier)

        return tier

    def remove_tier(self, index):
        """
        Remove a tier from the TextGrid.

        Parameters
        ----------
        tier : int
            The tier index.

        Returns
        -------
        :class:`IntervalTier` or :class:`PointTier`
            The removed tier.
        """
        return self._tiers.pop(index)

    def get_tier_by_name(self, tier_name):
        """
        Return a list of tier objects with a specific name.

        Parameters
        ----------
        tier : str
            The name of a tier.

        Returns
        -------
        list of class:`mytextgrid.core.tier.Tier`
            A list of tiers objects
        """
        if not isinstance(tier_name, str):
            raise TypeError('tier MUST BE a str')

        list_ = [tier for tier in self._tiers if tier.name == tier_name]
        return list_

    def to_dict(self):
        """
        Convert a TextGrid into a dict.
        """
        tiers_list = []
        for tier in self._tiers:
            # Collect Intervals or points into a list
            items_list = []
            for item in tier:
                if tier.is_interval():
                    item = {
                        'xmin':item.xmin,
                        'xmax':item.xmax,
                        'text':item.text
                    }
                else:
                    item = {
                        'number':item.time,
                        'mark':item.text
                    }
                items_list.append(item)

            # Put intervasl/points collection into tiers
            tiers_list.append(
                {
                    'interval_tier': tier.is_interval(),
                    'name': tier.name,
                    'items': items_list
                }
            )
        # Put tiers into TextGrid
        textgrid_dict = {
            'xmin': self._xmin,
            'xmax': self._xmax,
            'tiers': tiers_list
        }
        return textgrid_dict

    def write(self, path, short_format = False, encoding = 'utf-8'):
        """
        Write a TextGrid object as a text file.

        Parameters
        ----------
        path : str or :class:`pathlib.Path`
            The path where the TextGrid file will be written.
        short_format : bool, default False
            ``True`` to output TextGrid file in long format. ``False`` for short format.
        encoding : str, default 'utf-8'
            The encoding of the file.
        """
        textgrid_out.textgrid_to_text_file(self, path, short_format, encoding)

    def write_as_json(self, *args, **kwds):
        """
        Write a TextGrid object as a JSON file.

        Parameters
        ----------
        path : str or :class:`pathlib.Path`
            The path where the JSON file will be written.
        encoding : str, default utf-8
            The encoding of the resulting file.
        """
        textgrid_out.textgrid_to_json(self, *args, **kwds)



"""Create and manipulate point tiers objects"""
import decimal
from mytextgrid.core.tier import Tier
from mytextgrid.eval import obj_to_decimal
decimal.getcontext().prec = 16

class PointTier(Tier):
    """
    Represent a tier that contains Point objects.
    """
    def __init__(self, name = '', xmin = 0, xmax = 1):
        super().__init__(name, xmin, xmax, is_interval = False)

    def insert_point(self, time, text = ''):
        """
        Insert a Point into PointTier.

        Parameters
        ----------
        time : int, float, str or :class:`decimal.Decimal`
            The time in seconds where a Point object will be inserted.
        text : str
            The text of the selected Point.
        """
        time_ = obj_to_decimal(time)
        self.eval_time_range(time_) # Check if out of range

        index = self.get_index_at_time(time_)
        if not index is None:
            raise ValueError(f'Cannot insert a Point at {time}.')

        point = Point(time_, text)
        self._items.append(point)
        self._items.sort(key=lambda x:x.time)

    def insert_points(self, *times):
        """Insert various empty points into PointTier.

        Parameters
        ----------
        *times : float or decimal.Decimal
            One or more time items where Point items will be inserted.
        """
        for time in times:
            self.insert_point(time)

    def remove_point(self, index):
        """
        Remove a Point.

        Parameters
        ----------
        index : int
            The index of the Point in PointTier. It must be 0 <= index < len(PointTier).
        """
        self._items.pop(index)

    def get_index_at_time(self, time):
        """
        Get the index of an existing `Point`.

        Parameters
        ----------
        time : int, float, str or :class:`decimal.Decimal`
            The position in seconds of the :class:`mytextgrid.core.point_tier.Interval`.

        Returns
        -------
        int or None
            Return the index of the point. If not found, return `None`.
        """
        time_ = obj_to_decimal(time)
        self.eval_time_range(time_) # Check if out of range

        for index, point in enumerate(self._items):
            if point.time == time_:
                return index
        return None

    def get_point_at_time(self, time):
        """
        Get the point at the specified time in the PointTier.

        Parameters
        ----------
        time : int, float, str or :class:`decimal.Decimal`
            The position in seconds of the :class:`mytextgrid.core.point_tier.Interval`.

        Returns
        -------
        :class:`mytextgrid.core.point_tier.Point` or None
            Return the index of the point. If not found, return `None`.
        """
        index = self.get_index_at_time(time)
        if index is None:
            return None
        return self._items[index]

class Point:
    """Represent a Point object which is the minimal unit of a PointTier object"""

    def __init__(self, time, text = ''):

        # Check input type
        if not isinstance(time, (int, float, str, decimal.Decimal)):
            raise TypeError('time MUST BE an int, float, str or decimal.Decimal.')
        if not isinstance(text, str):
            raise TypeError('text MUST BE a str.')

        self._time = obj_to_decimal(time)
        self._text = text

    @property
    def time(self):
        """
        Return _time attribute.
        """
        return self._time

    @property
    def xmin(self):
        """
        Return _time attribute.
        """
        return self._time

    @property
    def xmax(self):
        """
        Return _time attribute.
        """
        return self._time

    @property
    def text(self):
        """
        Return _text attribute.
        """
        return self._text

    @text.setter
    def text(self, value):
        """
        Modify _text attribute.
        """
        if not isinstance(value, str):
            raise TypeError('text MUST BE a str')
        self._text = value



"""A tiny python project for creating, reading, writing and querying Praat annotation files
(TextGrid).
"""
from mytextgrid.core.textgrid import create_textgrid
from mytextgrid.io.textgrid_in import read_from_file
from mytextgrid.io.textgrid_in import read_from_stream
from mytextgrid.core.textgrid import TextGrid


import decimal
decimal.getcontext().prec = 16

def obj_to_decimal(time, message = None):
    """
    Convert a number to :class:`decimal.Decimal`.

    Parameters
    ----------
    time : int, float str or decimal.Decimal
        A number in seconds to be converted.
    message : str
        A message to show if raise an exception.

    Returns
    -------
    :class:`decimal.Decimal`
        A :class:`decimal.Decimal` object representing the time in seconds.
    """
    if isinstance(time, decimal.Decimal):
        return time
    elif isinstance(time, (int, float)):
        return decimal.Decimal(str(time))
    elif isinstance(time, str):
        return decimal.Decimal(time)
    else:
        message = 'time parameter must be int, float, str or decimal.Decimal'
        raise TypeError(message)



"""Create and manipulate tier objects"""
import decimal
from mytextgrid.eval import obj_to_decimal
decimal.getcontext().prec = 16

class Tier:
    """
    This is a base class for :class:`IntervalTier` and
    :class:`PointTier classes`. It represents an item's container.
    """
    def __init__(self, name = '', xmin = 0, xmax = 1, is_interval = True):
        """
        A base class to build a Interval or Point tier.

        Parameters
        ----------
        name : str, default ''
            The name of the tier.
        xmin : int, float str or decimal.Decimal
            The starting time (in seconds) of the tier.
        xmin : int, float str or decimal.Decimal
            The ending time (in seconds) of the tier.
        is_interval : bool, defaul True
            True is it is an Interval tier. False if it's a Point tier.
        """
        # Check input type
        if not isinstance(name, str):
            raise TypeError('name MUST BE a str.')
        if not isinstance(xmin, (int, float, str, decimal.Decimal)):
            raise TypeError('xmin MUST BE an int, float, str or decimal.Decimal.')
        if not isinstance(xmax, (int, float, str, decimal.Decimal)):
            raise TypeError('xmax MUST BE an int, float, str or decimal.Decimal.')
        if not isinstance(is_interval, bool):
            raise TypeError('is_interval MUST BE a boolean value.')

        xmin_ = obj_to_decimal(xmin)
        xmax_ = obj_to_decimal(xmax)

        assert xmax_ > xmin_, 'xmax MUST BE greater than xmin'

        self._name = name
        self._xmin = xmin_
        self._xmax = xmax_
        self._is_interval = is_interval
        self._items = []

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return iter(self._items)

    def __getitem__(self, key):
        return self._items[key]

    @property
    def name(self):
        """
        Return the `self._name` attribute.
        """
        return self._name

    @property
    def xmin(self):
        """
        Return the `self._xmin` attribute.
        """
        return self._xmin

    @property
    def xmax(self):
        """
        Return the `self._xmax` attribute.
        """
        return self._xmax

    @property
    def items(self):
        """
        Return the `self._items` attribute.
        """
        return self._items

    def get_duration(self):
        """
        Return the duration (in seconds) of the Tier instance.

        Return
        ------
            decimal.Decimal
                The duration in seconds of a Tier instance.
        """
        return self.xmax - self.xmin

    def is_interval(self):
        """"
        Check if the Tier contains interval or point items.

        Returns
        ------
        bool
            Return True if Tier contains intervals. Otherwise, return False.
        """
        return self._is_interval

    def eval_time_range(self, time):
        """
        Raise an exception if out of the tier range.
        """
        time_ = obj_to_decimal(time)

        if self._xmin > time_ > self._xmax:
            raise ValueError(f'{time_} (s) is out of range of the tier {self._name}')
        if self._xmin == time_:
            raise ValueError(f'{time_} seconds is at the left edge of the tier {self._name}.')
        if self._xmax == time_:
            raise ValueError(f'{time_} seconds is at the right edge of the tier {self._name}.')
